package ru.dbotthepony.mc.otm.client.screen.panels

import com.mojang.blaze3d.platform.InputConstants
import com.mojang.blaze3d.vertex.PoseStack
import it.unimi.dsi.fastutil.chars.CharOpenHashSet
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap
import net.minecraft.client.gui.screens.Screen
import ru.dbotthepony.mc.otm.client.isCtrlDown
import ru.dbotthepony.mc.otm.client.isKeyDown
import ru.dbotthepony.mc.otm.client.isShiftDown
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.client.render.DynamicBufferSource
import ru.dbotthepony.mc.otm.client.render.TextAlign
import ru.dbotthepony.mc.otm.client.render.drawAligned
import ru.dbotthepony.mc.otm.core.math.RGBAColor
import ru.dbotthepony.mc.otm.milliTime

open class TextInputPanel<out S : Screen>(
	screen: S,
	parent: EditablePanel<*>?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = 10f,
	height: Float = 10f,
) : EditablePanel<S>(screen, parent, x, y, width, height) {
	data class TextSelection(val start: Int, val end: Int) {
		init {
			require(start <= end) { "$start <= $end" }
			require(start >= 0) { "$start >= 0" }
		}
	}

	private inner class Snapshot {
		private val lines = ArrayList(this@TextInputPanel.lines) // ultra fast copy
		private val cursorLine = this@TextInputPanel.cursorLine
		private val cursorCharacter = this@TextInputPanel.cursorCharacter

		fun apply() {
			this@TextInputPanel.lines.clear()
			this@TextInputPanel.lines.addAll(lines)
			this@TextInputPanel.cursorCharacter = cursorCharacter
			this@TextInputPanel.cursorLine = cursorLine
		}

		override fun equals(other: Any?): Boolean {
			if (this === other) return true
			if (javaClass != other?.javaClass) return false

			other as TextInputPanel<*>.Snapshot

			if (lines != other.lines) return false
			if (cursorLine != other.cursorLine) return false
			if (cursorCharacter != other.cursorCharacter) return false

			return true
		}

		override fun hashCode(): Int {
			var result = lines.hashCode()
			result = 31 * result + cursorLine
			result = 31 * result + cursorCharacter
			return result
		}
	}

	protected var ignoreUpdates = false
	var debug = true
	var multiLine = false
	var cursorLine = 0
	var cursorCharacter = 0
	var textColor = RGBAColor.WHITE
	var cursorColor = RGBAColor.BLACK

	private var textCache: String? = null
	private val lines = ArrayList<String>()
	private val selections = Int2ObjectOpenHashMap<TextSelection>()
	private val undo = ArrayDeque<Snapshot>()
	private val redo = ArrayDeque<Snapshot>()

	private var snapshotTimer: Long? = null

	override fun tick() {
		super.tick()

		if (snapshotTimer != null && snapshotTimer!! <= milliTime) {
			pushbackSnapshot()
			snapshotTimer = null
			redo.clear()
		}
	}

	private fun pushbackSnapshot() {
		val snapshot = Snapshot()

		if (undo.isEmpty() || undo.last() != snapshot)
			undo.addLast(snapshot)
	}

	private fun recordHistory(hard: Boolean = false) {
		if (hard) {
			pushbackSnapshot()
			snapshotTimer = null
		} else if (snapshotTimer == null) {
			snapshotTimer = milliTime + 800L
		} else if (snapshotTimer!! <= milliTime) {
			pushbackSnapshot()
			snapshotTimer = null
		}

		redo.clear()
	}

	// it gets really dirty
	// idk man :(
	private fun undo() {
		if (undo.isEmpty())
			return

		if (snapshotTimer != null || redo.isEmpty()) {
			pushbackSnapshot()
			snapshotTimer = null
		}

		val current = Snapshot()

		while (undo.isNotEmpty()) {
			val removed = undo.removeLast()
			redo.addFirst(removed)

			if (removed != current) {
				removed.apply()
				break
			}
		}
	}

	// it gets really dirty
	// idk man :(
	private fun redo() {
		if (redo.isEmpty())
			return

		val current = Snapshot()

		snapshotTimer = null

		while (redo.isNotEmpty()) {
			val removed = redo.removeFirst()
			undo.addLast(removed)

			if (current != removed) {
				removed.apply()
				break
			}
		}
	}

	operator fun get(index: Int): String? {
		return lines.getOrNull(index)
	}

	operator fun set(index: Int, value: String) {
		if (index < 0) {
			throw IndexOutOfBoundsException("negative index $index")
		}

		lines.ensureCapacity(index)

		while (lines.size <= index) {
			lines.add("")
		}

		lines[index] = value
	}

	fun insertLine(index: Int, value: String = "") {
		lines.ensureCapacity(index)

		while (lines.size < index) {
			lines.add("")
		}

		val upperLines = selections.int2ObjectEntrySet().filter { it.intKey >= index }

		for (entry in upperLines) {
			selections.remove(entry.intKey)
		}

		for (entry in upperLines) {
			selections.put(entry.intKey + 1, entry.value)
		}

		lines.add(index, value)
	}

	fun removeLine(index: Int): Boolean {
		if (index < 0 || index >= lines.size) {
			return false
		}

		if (cursorLine == index) {
			if (index != 0)
				cursorLine--

			cursorCharacter = lines.getOrNull(cursorLine)?.length ?: 0
		}

		lines.removeAt(index)
		selections.remove(index)
		val upperLines = selections.int2ObjectEntrySet().filter { it.intKey > index }

		for (entry in upperLines) {
			selections.remove(entry.intKey)
		}

		for (entry in upperLines) {
			selections.put(entry.intKey - 1, entry.value)
		}

		return true
	}

	fun moveCursors(line: Int, character: Int, moveBy: Int) {
		@Suppress("name_shadowing")
		val moveBy = if (character + moveBy < 0) -character else moveBy

		if (cursorLine == line && cursorCharacter >= character) {
			cursorCharacter = (cursorCharacter + moveBy).coerceIn(0, this[cursorLine]?.length ?: Int.MAX_VALUE)
		}

		val selection = selections[line]

		if (selection != null) {
			val (start, end) = selection

			if (character < start) {
				selections[line] = TextSelection(start + moveBy, end + moveBy)
			} else if (character > end && character + moveBy < end) {
				if (character + moveBy < start) {
					// selections.remove(line)
					selections[line] = TextSelection(0, 0)
				} else {
					selections[line] = TextSelection(start, character + moveBy)
				}
			} else if (character in start .. end) {
				if (moveBy > 0 || character + moveBy <= start) {
					selections[line] = TextSelection(start, end + moveBy)
				} else {
					selections[line] = TextSelection(character + moveBy, end + moveBy)
				}
			}
		}
	}

	fun wipeSelection() {
		val inversed = ArrayList(selections.int2ObjectEntrySet())
		inversed.sortByDescending { it.intKey }

		for ((lineNumber, selection) in inversed) {
			if (selection.start != selection.end) {
				if (selection.start == 0 && selection.end == this[lineNumber]!!.length) {

				}
			}
		}

		selections.clear()
	}

	protected open fun textChanged(oldText: String, newText: String) {}

	var text: String
		get() {
			var textCache = textCache

			if (textCache == null) {
				textCache = lines.joinToString("\n")
				this.textCache = textCache
			}

			return textCache
		}
		set(value) {
			if ((textCache ?: lines.joinToString("\n")) == value)
				return

			snapshotTimer = null
			selections.clear()
			lines.clear()
			redo.clear()
			undo.clear()
			cursorLine = 0
			cursorCharacter = 0
			lines.addAll(value.split(NEWLINES))
		}

	fun advanceCursorLeft() {
		val line = this[cursorLine]

		if (line != null && cursorCharacter > line.length) {
			cursorCharacter = line.length - 1
		}

		if (cursorCharacter > 0) {
			cursorCharacter--
		} else if (cursorLine > 0) {
			cursorLine--
			cursorCharacter = 0

			@Suppress("name_shadowing")
			val line = this[cursorLine]

			if (line != null) {
				cursorCharacter = line.length
			}
		}
	}

	fun advanceCursorRight() {
		cursorCharacter++

		val line = this[cursorLine]

		if (line != null && cursorCharacter > line.length) {
			if (lines.size <= cursorLine + 1) {
				cursorCharacter = line.length
			} else {
				cursorLine++
				cursorCharacter = 0
			}
		} else if (line == null) {
			cursorCharacter = 0
		}
	}

	override fun keyPressedInternal(key: Int, scancode: Int, mods: Int): Boolean {
		if (key == InputConstants.KEY_ESCAPE) {
			killFocus()
			return true
		}

		if (key == InputConstants.KEY_RETURN) {
			if (multiLine) {
				if (!minecraft.window.isShiftDown && !minecraft.window.isCtrlDown)
					wipeSelection()

				val line = this[cursorLine]

				if (line != null && !minecraft.window.isShiftDown) {
					if (cursorCharacter <= 0) {
						recordHistory(true)
						insertLine(cursorLine, "")

						if (!minecraft.window.isCtrlDown) {
							cursorLine++
							cursorCharacter = 0
						}
					} else if (cursorCharacter >= line.length) {
						recordHistory(true)
						insertLine(cursorLine + 1, "")

						if (!minecraft.window.isCtrlDown) {
							cursorLine++
							cursorCharacter = 0
						}
					} else {
						val before = line.substring(0, cursorCharacter)
						val after = line.substring(cursorCharacter)

						recordHistory(true)
						this[cursorLine] = before
						insertLine(cursorLine + 1, after)

						if (!minecraft.window.isCtrlDown) {
							cursorLine++
							cursorCharacter = 0
						}
					}
				} else {
					recordHistory(true)
					insertLine(cursorLine + 1)
					cursorLine++
					cursorCharacter = 0
				}

				return true
			} else {
				killFocus()
				return true
			}
		}

		if (key == InputConstants.KEY_TAB) {
			if (multiLine) {
				val lines = if (selections.isNotEmpty()) selections.keys else listOf(cursorLine)


				if (minecraft.window.isKeyDown(InputConstants.KEY_RSHIFT) || minecraft.window.isKeyDown(InputConstants.KEY_LSHIFT)) {
					var hit = false

					for (lineNumber in lines) {
						val line = this[lineNumber]!!

						if (line[0] == '\t') {
							if (!hit) {
								hit = true
								recordHistory(true)
							}

							this[lineNumber] = line.substring(1)
							moveCursors(lineNumber, 1, -1)
						}
					}
				} else {
					if (lines.isNotEmpty())
						recordHistory(true)

					for (lineNumber in lines) {
						this[lineNumber] = "\t" + this[lineNumber]!!
						moveCursors(lineNumber, 0, 1)
					}
				}

				return true
			}

			killFocus()
			return true
		}

		if (key == InputConstants.KEY_LEFT) {
			if (minecraft.window.isCtrlDown) {
				val line = this[cursorLine]

				if (line == null || cursorCharacter <= 0) {
					advanceCursorLeft()
				} else {
					if (cursorCharacter >= line.length)
						cursorCharacter = line.length - 1

					cursorCharacter = greedyAdvanceLeft(line, cursorCharacter)
				}
			} else {
				advanceCursorLeft()
			}

			return true
		} else if (key == InputConstants.KEY_RIGHT) {
			if (minecraft.window.isCtrlDown) {
				val line = this[cursorLine]

				if (line == null || cursorCharacter + 1 >= line.length) {
					advanceCursorRight()
				} else {
					if (cursorCharacter < 0)
						cursorCharacter = 0

					cursorCharacter = greedyAdvanceRight(line, cursorCharacter)
				}
			} else {
				advanceCursorRight()
			}

			return true
		} else if (key == InputConstants.KEY_UP) {
			if (cursorLine > 0) {
				cursorLine--
			}

			return true
		} else if (key == InputConstants.KEY_DOWN) {
			if (cursorLine < lines.size - 1) {
				cursorLine++
			}

			return true
		}

		if (key == InputConstants.KEY_BACKSPACE) {
			if (cursorLine <= 0 && cursorCharacter <= 0) {
				return true
			} else if (cursorCharacter <= 0) {
				val line = this[cursorLine]

				if (line == null) {
					cursorCharacter = this[--cursorLine]?.length ?: 0
				} else {
					removeLine(cursorLine)
					val newLine = this[cursorLine]

					if (newLine == null) {
						this[cursorLine] = line
					} else {
						this[cursorLine] = newLine + line
					}

					recordHistory()
				}
			} else {
				val line = this[cursorLine]

				if (line != null) {
					var cursorCharacter = cursorCharacter

					if (cursorCharacter >= line.length)
						cursorCharacter = line.length - 1

					val newLine = line.substring(0, cursorCharacter - 1) + line.substring(cursorCharacter)
					moveCursors(cursorLine, cursorCharacter, -1)
					this[cursorLine] = newLine
					recordHistory()
				}
			}

			return true
		}

		if (key == InputConstants.KEY_DELETE) {
			if (cursorLine !in 0 until lines.size) {
				cursorLine = lines.size - 1
				return true
			}

			if (minecraft.window.isShiftDown) {
				pushbackSnapshot()
				removeLine(cursorLine)

				if (cursorLine >= lines.size)
					cursorLine = lines.size

				cursorCharacter = 0
				recordHistory(true)
				return true
			}

			val line = this[cursorLine]!!

			if (cursorCharacter >= line.length) {
				if (cursorLine + 1 == lines.size) {
					return true
				}

				if (snapshotTimer == null)
					pushbackSnapshot()

				val bottomLine = this[cursorLine + 1]!!
				cursorCharacter = line.length
				this[cursorLine] = line + bottomLine
				removeLine(cursorLine + 1)

				recordHistory()
			} else {
				if (snapshotTimer == null)
					pushbackSnapshot()

				val cursorCharacter = cursorCharacter
				moveCursors(cursorLine, cursorCharacter, -1)
				this[cursorLine] = line.substring(0, cursorCharacter) + line.substring(cursorCharacter + 1)

				if (cursorCharacter != 0)
					this.cursorCharacter++

				recordHistory()
			}

			return true
		}

		if (key == InputConstants.KEY_Z && minecraft.window.isCtrlDown) {
			undo()
		} else if (key == InputConstants.KEY_Y && minecraft.window.isCtrlDown) {
			redo()
		}

		return true
	}

	override fun charTypedInternal(codepoint: Char, mods: Int): Boolean {
		var line = this[cursorLine]

		if (line == null) {
			set(cursorLine, "")
			line = ""
			cursorCharacter = 0
		}

		if (cursorCharacter >= line.length)
			line += codepoint
		else
			line = line.substring(0, cursorCharacter) + codepoint + line.substring(cursorCharacter)

		if (snapshotTimer == null)
			pushbackSnapshot()

		set(cursorLine, line)
		moveCursors(cursorLine, cursorCharacter, 1)
		recordHistory()

		return true
	}

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		var y = 0f

		for (line in lines) {
			font.drawAligned(
				poseStack = stack,
				buffer = BUFFER,
				text = line,
				align = TextAlign.TOP_LEFT,
				x = 0f,
				y = y,
				color = textColor
			)

			y += font.lineHeight + 2f
		}

		if (isFocused && milliTime % 1000L > 200L) {
			val activeLine = this[cursorLine]

			if (activeLine == null || cursorCharacter >= activeLine.length) {
				font.drawAligned(
					poseStack = stack,
					buffer = BUFFER,
					text = "_",
					align = TextAlign.TOP_LEFT,
					x = if (activeLine == null) 0f else font.width(activeLine).toFloat(),
					y = cursorLine * (font.lineHeight + 2f),
					color = cursorColor
				)
			} else {
				font.drawAligned(
					poseStack = stack,
					buffer = BUFFER,
					text = "|",
					align = TextAlign.TOP_LEFT,
					x = font.width(activeLine.substring(0, cursorCharacter)).toFloat() - 1f,
					y = cursorLine * (font.lineHeight + 2f),
					color = cursorColor
				)
			}
		}

		if (debug) {
			font.drawAligned(
				poseStack = stack,
				buffer = BUFFER,
				text = cursorLine.toString(),
				align = TextAlign.TOP_RIGHT,
				x = width,
				y = 0f,
				color = cursorColor
			)

			font.drawAligned(
				poseStack = stack,
				buffer = BUFFER,
				text = cursorCharacter.toString(),
				align = TextAlign.TOP_RIGHT,
				x = width,
				y = font.lineHeight + 2f,
				color = cursorColor
			)

			font.drawAligned(
				poseStack = stack,
				buffer = BUFFER,
				text = lines.size.toString(),
				align = TextAlign.TOP_RIGHT,
				x = width,
				y = font.lineHeight * 2 + 4f,
				color = cursorColor
			)
		}

		BUFFER.endBatch()
	}

	override fun mouseClickedInner(x: Double, y: Double, button: Int): Boolean {
		requestFocus()
		return true
	}

	private enum class CharType {
		SPACES {
			override fun contains(input: Char): Boolean {
				return _SPACES.contains(input)
			}

			override fun canJumpInto(category: CharType): Boolean {
				return true
			}
		},
		CONTROL {
			override fun contains(input: Char): Boolean {
				return _CONTROL.contains(input)
			}

			override fun canJumpInto(category: CharType): Boolean {
				return category == CONTROL
			}
		},
		OTHER {
			override fun contains(input: Char): Boolean {
				return !_SPACES.contains(input) && !_CONTROL.contains(input)
			}

			override fun canJumpInto(category: CharType): Boolean {
				return category == OTHER
			}
		};

		abstract fun contains(input: Char): Boolean
		abstract fun canJumpInto(category: CharType): Boolean

		companion object {
			private val values = values()

			operator fun get(input: Char): CharType {
				for (value in values) {
					if (value.contains(input)) {
						return value
					}
				}

				throw NoSuchElementException("for $input")
			}
		}
	}

	companion object {
		val NEWLINES = Regex("\n\r?")
		private val BUFFER = DynamicBufferSource()

		private val _SPACES = CharOpenHashSet().also {
			for (char in " \t\n\r\u0000")
				it.add(char)
		}

		private val _CONTROL = CharOpenHashSet().also {
			for (char in "!@\"'#$%^&?*()~`[]/\\;:|<>.,")
				it.add(char)
		}

		private fun greedyAdvanceLeft(input: String, position: Int): Int {
			if (position <= 1)
				return 0

			@Suppress("name_shadowing")
			var position = position
			var type = CharType[input[--position]]

			while (position > 0) {
				val newType = CharType[input[position]]

				if (type.canJumpInto(newType)) {
					position--
					type = newType
				} else {
					return position + 1
				}
			}

			return position
		}

		private fun greedyAdvanceRight(input: String, position: Int): Int {
			@Suppress("name_shadowing")
			var position = position
			var type = CharType[input[position]]

			while (position < input.length) {
				val newType = CharType[input[position]]

				if (type.canJumpInto(newType)) {
					position++
					type = newType
				} else {
					return position
				}
			}

			return position
		}
	}
}
